package WordNet::IO::Cache;

# HACK(ish): -- might update later...
{ no strict 'refs'; *{$_} = use_underlying($_) for qw/sensenumber sk_to_sn sk_to_sensenumber count sentenceids sentences/; }
# END HACK

use 5.006;
use strict;
use warnings;
use Carp;

use base 'WordNet::IO';
use WordNet::Filenames;
use WordNet qw/:vars/;

our (%EXPORT_TAGS, @EXPORT_OK, @EXPORT);
%EXPORT_TAGS = @EXPORT_OK = @EXPORT = ();

our $VERSION = '0.01';

my @levels = qw/none is_word ids index data/;
my %levels = map { ($_, $levels[$_]); } (0..$#levels);
foreach (values %levels) { $levels{$_} = $_; }

our @defaults = ( underlying => 'Disk',
                  mechanism  => (eval { require 5.008; 'Storable' }
                                 or 'Data::Dumper'),
                  cachedir   => '.',
                  cachelevel => 'is_word',
                  compact    => 1,
                  identify   => 'size' );

sub new {
    my $class = shift;
    my $self = WordNet::IO::new($class, @defaults, @_);
    croak "Invalid cachelevel: $self->{cachelevel}"
        unless defined $levels{$self->{cachelevel}};
    eval "require $self->{mechanism}"
        or croak "Could not load: $self->{mechanism}\n$!\n$@";
    eval "require WordNet::IO::$self->{underlying}"
        or croak "Can't use underlying method: $self->{underlying}\nMaybe you need to create WordNet::IO::$self->{underlying}?";
    my $reader = "WordNet\::IO\::$self->{underlying}"->new(@defaults, @_)
        or croak "Couldn't create a new WordNet\::IO\::$self->{underlying} object";
    return $reader if $reader->{no_cache};
    $self->{cacheREADER} = $reader;
    $self->{FILENAMES} = WordNet::Filenames->new(@defaults, @_);
    $self->identify();
    $self->load_cached();
    return $self;
}

sub filenames { shift->{FILENAMES} }

my %handle = ( none    => [],
               is_word => [ 'is_word' ],
               ids     => [ 'id', 'id_all', 'id_by_sk' ],
               index   => [ 'iterator' ],
               data    => [ 'read' ] );

sub use_underlying {
    my $sub = shift;
    return sub { shift->{cacheREADER}->$sub(@_); }
}

sub identify {
    my $self = shift;
    my $method = $self->{identify};
    my $identity = '';
	my $filenames = $self->filenames;
	my @files = sort map { $filenames->data($_), $filenames->index($_) } @pos;
    if ($method eq 'size') {
        $identity = 0;
        $identity += -s $_ for @files;
    } elsif ($method eq 'MD5') {
        require Digest::MD5;
        my $digest = Digest::MD5->new;
        for (@files) {
            open MD5, '<', $_ or warn "Couldn't open $_ for reading" and next;
            $digest->addfile(*MD5);
            close MD5;
        }
        $identity = $digest->hexdigest;
    } elsif ($method eq 'date') {
        my $file = $files[0];
        $file = -M $_ < -M $file ? $_ : $file for @files;
        $identity = (stat $file)[9]; # mtime of most recently mod'ed file
    } else {
        die "Invalid identification scheme: $method";
    }
    $self->{identity} = $identity;
}

sub load_cached {
    my $self = shift;
    foreach my $level (@levels) {
        my $part = $levels{$level};
        unless ($self->load_part($part) || $self->create_part($part)) {
            carp "Couldn't load $part in $self->{class}\::load_cached";
        }
        last if $level eq $self->{cachelevel};
    }
    my $high_enough = 0;
    foreach my $level (@levels) {
        $high_enough = 1 if $level eq $self->{cachelevel};
        next unless $high_enough++ > 1;
        {
            no strict 'refs';
            foreach my $sub (@{$handle{$level}}) {
                *{"$self->{class}\::$sub"} = use_underlying $sub;
            }
        }
    }
}

my %files = ( none    => [],
              is_word => [ 'is_word',
                           'is_word_lower' ],
              ids     => [ 'ids',       'ids.rev',       'senses',
                           'ids_lower', 'ids_lower.rev', 'senses_lower' ],
              index   => [ 'index' ],
              data    => [ 'data' ] );
my %abbrev = ( 'Data::Dumper' => 'dmp',
               'Storable'     => 'bin' );

sub load_part {
    my $self = shift;
    my ($part) = @_;
    foreach my $file (@{$files{$part}}) {
        my $id = $self->{identity} || '';
        my $filename = $self->filenames->catfile('cache', $id.$file);
        my $actual = "$filename.$abbrev{$self->{mechanism}}";
        eval {
            my $t = time;
            -f $actual and -r _ and utime $t, $t, $actual;
            1;
        } or warn $actual.$!.$@;
        return 0 unless ($self->_retrieve($filename, $file));
    }
    return 1;
}

my %load_sub = ( none    => sub { },
                 is_word => \&load_words,
                 ids     => \&load_ids,
                 index   => \&load_index,
                 data    => \&load_data );

sub create_part {
    my $self = shift;
    my ($part) = @_;
    $load_sub{$part}->($self) or return 0;
    foreach my $file (@{$files{$part}}) {
        my $id = $self->{identity} || '';
        my $filename = $self->filenames->catfile('cache', $id.$file);
        $self->_store($filename, $file)
            or carp "Unable to store $filename: $! $@";
    }
    return 1;
}

sub _store {
    my $self = shift;
    my ($file, $hash_key) = @_;
    my $mech = $self->{mechanism};
    $file = "$file.$abbrev{$mech}";

    if ($mech eq 'Data::Dumper') {
        open my $fh, ">$file" or return 0;
        print $fh $mech->new([$self->{$hash_key}],
                             ["\$self->{".$hash_key."}"])->Indent(0)->Dump;
        close $fh;
    } elsif ($mech eq 'Storable') {
		eval { Storable::nstore($self->{$hash_key}, $file); } or return 0;
    } else {
        carp "Cache mechanism $mech not found in ".__PACKAGE__."\::load_hash";
        return 0;
    }
    return 1;
}

sub _retrieve {
    my $self = shift;
    my ($file, $hash_key) = @_;
    my $mech = $self->{mechanism};
    $file = "$file.$abbrev{$mech}";

    if ($mech eq 'Data::Dumper') {
        eval { $self->{$hash_key} = do "$file"; } or return 0;
    } elsif ($mech eq 'Storable') {
        eval "\$self->{$hash_key} = $mech\::retrieve(\$file);" or return 0;
    } else {
        carp "Cache mechanism $mech not found in $self->{class}\::_retrieve";
        return 0;
    }
    return 1;
}

sub load_words {
    my $self = shift;
    $self->{is_word} = $self->{cacheREADER}->is_word_hash;
    $self->{is_word_lower} = {};
    my $hash = $self->{is_word_lower};
    foreach my $key (keys %{$self->{is_word}}) {
        $hash->{lc($key)} += $self->{is_word}{$key};
    }
    return 1;
}

sub load_ids {
    return 0;
}

sub load_index {
    my $self = shift;
    carp "This level is not implemented: 'index' in $self->{class}";
    return 1;
}

sub load_data {
    my $self = shift;
    carp "This level is not implemented: 'data' in $self->{class}";
    return 1;
}

sub is_word {
    my ($self, $word, $pos) = @_;
    $word ||= '';
    $pos  ||= '';
    unless ($pos) { $self->is_word($word, $_) and return 1 for @pos; return 0 }
    $word =~ tr/_/ /;
    my $key = "$pos:$word";
    $key = lc $key unless $self->{case_sensitive};
    my $hash = $self->{$self->{case_sensitive} ? 'is_word' : 'is_word_lower'};
    return (defined $hash->{$key} ? $hash->{$key} : 0);
}

1;

package WordNet::IO::NONE;
use strict;
use warnings;
require Exporter;
our @ISA=qw(Exporter WordNet::IO);
sub new    { return bless {}, shift; }
sub id     { undef; }
sub id_all { (); }
sub read   { undef; }
sub write  { undef; }
1;

__END__
# Below is stub documentation for your module. You better edit it!

=head1 NAME

WordNet::IO::Cache - Perl extension for blah blah blah

=head1 SYNOPSIS

  use WordNet::IO::Cache;
  blah blah blah

=head1 DESCRIPTION

Stub documentation for WordNet::IO::Cache, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.


=head1 AUTHOR

Ben Haskell, E<lt>ben@clarity.princeton.eduE<gt>

=head1 SEE ALSO

L<perl>.

=cut
