package WordNet::IO::Disk;

use 5.006;
use strict;
use warnings;

use Search::Dict;

use WordNet qw/:pos/;
use base 'WordNet::IO';
use WordNet::Filenames;
use Scalar::Util qw/weaken/;

our (@EXPORT, @EXPORT_OK, %EXPORT_TAGS);
@EXPORT = @EXPORT_OK = %EXPORT_TAGS = ();
our $VERSION = '0.01';

sub new {
    my $class = shift;
    my $filenames = WordNet::Filenames->new(@_);

    my ($DATA_FILE, $IDX_FILE) = ({}, {});
    $DATA_FILE->{$_} = $filenames->data($_)  for @pos;
    $IDX_FILE->{$_}  = $filenames->index($_) for @pos;
    my $ID_FILE = $filenames->index('sense');

    my ($SIDX_FILE, $STXT_FILE, $FRM_FILE) =
        map $filenames->find($_),
    qw/sentidx.vrb sents.vrb frames.vrb/;

    my $NAME_FILE = $filenames->find(qw/lexnames filenames/);

	my %filenames = (
		id      =>   $ID_FILE,
		index   =>  $IDX_FILE,
		data    => $DATA_FILE,
		sentidx => $SIDX_FILE,
		sents   => $STXT_FILE,
		frames  =>  $FRM_FILE,
		names   => $NAME_FILE,
	);
	@_ = map @_[$_,$_+1], grep { not
		$_[$_] eq 'filenames'
		? (%filenames = (%filenames, %{$_[$_+1]})) # combine the filename hashes
		: 0
	} grep !($_ % 2), 0..$#_;
    my $self = WordNet::IO::new(
		$class,
		filenames => {(%filenames)},
		FILENAMES => $filenames,
		@_ );

    my %open_success = (); # MODPERL requires initialization
    for ( map([ 1, index => $_ ], @pos),
          map([ 1, data  => $_ ], @pos),
          [ 1, qw/id/ ],
          [ 0, qw/names/ ],
          [ 0, qw/sentidx/ ],
          [ 0, qw/sents/ ],
          [ 0, qw/frames/ ],
          ) {
        my ($critical, @keys) = @$_;
		my $fn = $self->{filenames};
		$fn = $fn->{$_} for @keys;
		unless ($fn) {
			my $mess = "Couldn't determine (@keys) filename.";
			my @found = grep $open_success{$_}, keys %open_success;
			$mess .= join "\n", "Other files were found at:", @found if @found;
			die $mess if $critical;
			next;
		}
		my ($fh) = ($self->{files} ||= {});
		($fh) = ($fh->{$_} ||= {}) for @keys[0..$#keys-1];
		if (open $fh->{$keys[-1]}, '<', $fn) {
			$open_success{$fn} = 1;
		} else {
			my $mess = "Couldn't open $fn: $!";
			my @found = grep $open_success{$_}, keys %open_success;
			$mess .= join "\n", "Other files were found at:", @found if @found;
			die $mess if $critical;
			warn $mess unless $self->{nowarnings};
		}
    }
    for ( [ qw/NOID     id           / ],
          [ qw/NOSENTS  sentidx sents/ ],
          ) {
        my ($flag, @files) = @$_;
        $self->{$flag} = 1 if grep { !$open_success{$self->{filenames}{$_}} } @files;
    }
    $self->_init_lexfilenames;
    $self->_init_frametext;
    if (0) { # This stuff is _REALLY_ slow -- might Cache it later
        $self->_init_countlist unless $self->{NOID};
        $self->_init_sentences unless $self->{NOSENTS};
    }
# map some 's'attelite stuff to the 'a'djective stuff
    $self->{files}{$_}{s} = $self->{files}{$_}{a} for qw/index data/;

    return $self;
}

our @lexfiles;
BEGIN { 
	@lexfiles = qw/00 adj.all 01 adj.pert 02 adv.all 03 noun.Tops
		04 noun.act 05 noun.animal 06 noun.artifact 07 noun.attribute
		08 noun.body 09 noun.cognition 10 noun.communication 11 noun.event
		12 noun.feeling 13 noun.food 14 noun.group 15 noun.location
		16 noun.motive 17 noun.object 18 noun.person 19 noun.phenomenon
		20 noun.plant 21 noun.possession 22 noun.process 23 noun.quantity
		24 noun.relation 25 noun.shape 26 noun.state 27 noun.substance
		28 noun.time 29 verb.body 30 verb.change 31 verb.cognition
		32 verb.communication 33 verb.competition 34 verb.consumption
		35 verb.contact 36 verb.creation 37 verb.emotion 38 verb.motion
		39 verb.perception 40 verb.possession 41 verb.social
		42 verb.stative 43 verb.weather 44 adj.ppl/;
}
sub _init_lexfilenames {
    my $self = shift;
    my $fh = $self->{files}{names};
    $self->{lexfilename} = {(@lexfiles)} and return unless $fh;
    local $_;
    while (<$fh>) {
        chomp;
        my ($n, $name, $p) = split;
        $self->{lexfilename}{$n} = $name;
    }
}

sub _init_countlist {
    my $self = shift;
    my $fh = $self->{files}{id};
    local $_;
    while (<$fh>) {
        chomp;
        my ($sk, $off, $sn, $count) = split;
        $self->{sensenumber}{$sk} = $sn;
        $self->{count}{$sk} = $count if $count;
    }
}

sub _init_sentences {
    my $self = shift;
    my ($idx, $txt) = map { $self->{files}{$_} } qw/sentidx sents/;
    local $_;
    while (<$idx>) {
        chomp;
        my ($sk, $sents) = split /\s/, $_, 2;
        $self->{sentenceidx}{$sk} = [ split /,/, $sents ];
    }
    while (<$txt>) {
        chomp;
        my ($id, $sprintf) = split /\s/, $_, 2;
        $self->{sentences}{$id} = $sprintf;
    }
}

sub _init_frametext {
    my $self = shift;
    my $fh = $self->{files}{frames};
    unless ($fh) {
        $self->{frames} = {
1  => 'Something ----s',
2  => 'Somebody ----s',
3  => 'It is ----ing',
4  => 'Something is ----ing PP',
5  => 'Something ----s something Adjective/Noun',
6  => 'Something ----s Adjective/Noun',
7  => 'Somebody ----s Adjective',
8  => 'Somebody ----s something',
9  => 'Somebody ----s somebody',
10 => 'Something ----s somebody',
11 => 'Something ----s something',
12 => 'Something ----s to somebody',
13 => 'Somebody ----s on something',
14 => 'Somebody ----s somebody something',
15 => 'Somebody ----s something to somebody',
16 => 'Somebody ----s something from somebody',
17 => 'Somebody ----s somebody with something',
18 => 'Somebody ----s somebody of something',
19 => 'Somebody ----s something on somebody',
20 => 'Somebody ----s somebody PP',
21 => 'Somebody ----s something PP',
22 => 'Somebody ----s PP',
23 => 'Somebody\'s (body part) ----s',
24 => 'Somebody ----s somebody to INFINITIVE',
25 => 'Somebody ----s somebody INFINITIVE',
26 => 'Somebody ----s that CLAUSE',
27 => 'Somebody ----s to somebody',
28 => 'Somebody ----s to INFINITIVE',
29 => 'Somebody ----s whether INFINITIVE',
30 => 'Somebody ----s somebody into V-ing something',
31 => 'Somebody ----s something with something',
32 => 'Somebody ----s INFINITIVE',
33 => 'Somebody ----s VERB-ing',
34 => 'It ----s that CLAUSE',
35 => 'Something ----s INFINITIVE',
};
        return;
    }
    local $_;
    while (<$fh>) {
        chomp;
        my ($id, $text) = split /\s/, $_, 2;
        $self->{frames}{$id} = $text;
    }
}

sub _look_it_up {
    my ($self, $fh, $sk) = @_;
    $sk =~ s/\%3:(?=\d\d:\d\d:[^:])/\%5:/ if $self->{conflate35};
    seek $fh, 0, 0;
    look $fh, $sk, 0, 0;
    local $_ = scalar <$fh>;
    return unless $_;
    my ($found) = split /\s/;
    return unless $found eq $sk;
    return $_;
}

sub filenames { shift->{FILENAMES} }

sub sensenumber {
    my ($s, $sk) = @_;
    return if $s->{NOID};
    return unless local $_ = $s->_look_it_up($s->{files}{id}, $sk);
    (split /\s/)[2];
}
*sk_to_sn = \&sensenumber;
*sk_to_sensenumber = \&sensenumber;

sub count {
    my ($s, $sk) = @_;
    return if $s->{NOID};
    return unless local $_ = $s->_look_it_up($s->{files}{id}, $sk);
    (split /\s/)[3];
}
sub sentenceids {
    my ($s, $sk) = @_;
    return if $s->{NOSENTS};
    return unless local $_ = $s->_look_it_up($s->{files}{sentidx}, $sk);
    split /,/, (split /\s/)[1];
}
sub sentences {
    my ($s, $sk, $w) = @_;
    return if $s->{NOSENTS};
    my $sf = $s->{files}{sents};
    grep { $_ } map {
        seek $sf, 0, 0;
        look $sf, $_, 0, 0;
        chomp(my $l = scalar(<$sf>) || '');
        my ($id, $f) = split /\s/, $l, 2;
        $id eq $_ ? sprintf($f, $w) : ();
    } $s->sentenceids($sk);
}

sub read {
    my $self = shift;
    my ($id, $just_words) = @_;
    $just_words = 0 if not defined $just_words;
    my ($pos, $off, $word) = split /:/, $id;
    my $fh = $self->{files}{data}{$pos};
    seek $fh, $off, 0;
    return [$self->_parse_line(scalar <$fh>, $just_words)] if $just_words;
    my $synset = $self->_parse_line(scalar <$fh>);
    return $synset->{words}[$word-1] if $word;
    return $synset;
}

sub id {
    my $self = shift;
    my ($word, $pos, $sense) = @_;
    $pos = $pos_short{$pos};
    $word = lc $word;
    $word =~ tr/ /_/;
    my $fh = $self->{files}{index}{$pos};

    my @ret = ();
    my $hyphen = $word;
    push @ret, $self->id($hyphen, $pos, $sense)
        if $hyphen =~ tr/_/-/;
    push @ret, $self->id($hyphen, $pos, $sense)
        if $hyphen =~ tr/-_ //d;
    seek $fh, 0, 0;
    look $fh, $word, 0, 0;
    my $line = undef; # MODPERL requires initialization
    return () unless defined($line = <$fh>);
    chomp $line;
    my @found = split /\s+/, $line;
    my ($fword, $fpos, $fcount) = @found;
    return @ret if $fword  ne $word;
    return @ret if $fpos   ne $pos;
    return @ret if $sense and $fcount  < $sense;
    return @ret if $sense and $sense   < 0;
    my (@offs) = splice @found, -$fcount;
    return "$pos\:$offs[$sense-1]" if ($sense);
    @ret = map { "$pos\:$_"; } @offs;
    return @ret;
}

sub id_all {
    my $self = shift;

    my @ret = ();
    foreach my $pos (@pos) {
        my $fh = $self->{files}{data}{$pos};
        seek $fh, 0, 0;
        local $_;
        while (<$fh>) {
            next unless /^(\d{8})/;
            push @ret, "$pos:$1";
        }
    }
    @ret;
}

use Getopt::Long;
sub iterator {
    my $parent = shift;
	my $options = Getopt::Long::Parser->new;
	$options->configure(pass_through => 'prefix_pattern=|-');
    my @parts = (); # MODPERL requires initialization
	my %p = ();
	my @counter_opts = ();
	local @ARGV = @_;
	$options->getoptions(
		map(($_ => \$p{$_}), keys %pos_short),
		'pos|part' => \@parts,
		'parts=s' => sub { push @parts, split /,/, $_[1]||''; },
		'dieat|die_at=i' => \ (my $die_at = 0),
		'<>' => sub {
			if ($_[0] =~ /^wc(.*)$/) { push @counter_opts, 'counter', $1; }
			elsif ($pos_short{$_[0]}) { warn "Intercepted: $_[0]\n"; push @parts, @_; }
			else { push @counter_opts, @_; }
		},
	);
	push @parts, grep defined($p{$_}), keys %p;
	my $counter = $parent->{WN}->counter(@counter_opts);
    @parts = @pos unless @parts;
	for (reverse 0..$#parts) {
		if (my $p = $pos_short{$parts[$_]}) {
			$parts[$_] = $p;
		} else {
			warn "Bad part of speech: $parts[$_]";
			splice @parts, $_, 1;
		}
	}
    my $ref = {
		files => [ map {
			open my $fh, '<', $parent->{filenames}{data}{$_}
				or die "Couldn't open file for iteration";
			($fh);
		} @parts ],
		used  => [],
		IO    => $parent,
		filepos => 0,
	};
    my $c = 0;
	$ref->{COUNT} = sub { $c; };
    $ref->{NEXT} = sub {
        my $self = shift;
        while (@{$self->{files}}) {
            my $fh = $self->{files}[0];
            seek $fh, $self->{filepos}, 0;
            local $_;
            while (<$fh>) {
                next unless /^\d/;
                my $ss = _parse_line($self->{IO}, $_);
                if ($ss) {
                    $self->{filepos} = tell $fh;
					$c = $counter->();
                    return ($die_at and $c > $die_at) ? undef : $ss;
                }
            }
            push @{$self->{used}}, shift @{$self->{files}};
            $self->{filepos} = 0;
        }
		$c = $counter->();
        return undef;
    };
    $ref->{RESET} = sub {
        my $self = shift;
        $self->{files}   = [ @{$self->{used}}, @{$self->{files}} ];
        $self->{used}    = [];
        $self->{filepos} = 0;
        seek $_, 0, 0 for @{$self->{files}};
    };
    WordNet::Iterator->new(%$ref);
}

my %convert        = qw/1 n 2 v 3 a 4 r 5 s/;
my %conflate       = qw/1 n 2 v 3 a 4 r 5 a/;
my %convert_back   = qw/n 1 v 2 a 3 r 4 s 5/;
my %conflate_back  = qw/n 1 v 2 a 3 r 4 s 3/;

sub id_by_sk {
    my $self = shift;
    my ($sk) = @_;
    my $line = $self->_look_it_up($self->{files}{id}, $sk);
    return unless $line;
    my ($found, $off) = split /\s/, $line;
    return unless $sk =~ /\%([1-5]):/;
    my $pos = ($self->{conflate35} ? $conflate{$1} : $convert{$1});
    return join ':', $pos, $off;
}

sub id_by_qd {
    my $self = shift;
    my ($qd) = @_;
    my ($w, $p, $s) = split /\#/, $qd;
    unless ($w and $p and $s) { warn "Bad QueryData ID: $qd\n"; return; }
    $w = lc($w);
    my $line = $self->_look_it_up($self->{files}{index}{$p}, $w);
    return unless $line;
    my (@dat) = split /\s/, $line;
    my ($pos, $n) = @dat[1,2];
    @dat = @dat[$#dat-$n+1..$#dat];
    return if $s > @dat or $s < 1;
    return join ':', $pos, $dat[$s-1];
}

sub sk_by_id {
    my $self = shift;
    my ($id, $word) = @_;

    my $ss = $self->read($id);
    return undef if not $ss;
    my $headword = $ss->wordnet_form($word, 1);
    my $sk = $headword->word;
    $sk =~ tr/ /_/;
    $sk = lc $sk . '%';
    $sk .= ($self->{conflate35}
            ? $conflate_back{$headword->pos}
            : $convert_back{$headword->pos});
    $sk .= ':';
    $sk .= sprintf('%02d', $ss->{lexfile}) . ':';
    $sk .= sprintf('%02d', $headword->{lexsense}) . ':';
    return "$sk:" unless $ss->{pos} =~ /[sa]/;
    return "$sk:" if $ss->{words}[0]->pointers('!');
    my @ptrs = $ss->pointers('&', 1);
    return "$sk:" unless @ptrs;
    my $headss = (shift @ptrs)->follow;
    return "$sk:" unless $headss;
    $headword = $headss->wordnet_form(undef, 1);
    $sk .= lc $headss->wordnet_form . ':';
    $sk .= sprintf('%02d', $headword->{lexsense});
    $sk;
}

sub is_word {
    my ($self, $word, $pos) = @_;
    unless ($pos) {
        $self->is_word($word, $_) and return 1 for @pos;
        return 0;
    }
    $word =~ tr/_/ /;
    $word = lc $word unless $self->{case_sensitive};
    for my $id ($self->id($word, $pos, 0)) {
        for my $w (map { $_->word } @{$self->read($id, 1)}) {
            $w = lc $w unless $self->{case_sensitive};
            return 1 if $w eq $word;
        }
    }
    0;
}

sub is_word_hash {
    my $self = shift;
    my $hash = {};
    foreach my $key (keys %{$self->{files}{data}}) {
        my $file = $self->{files}{data}{$key};
        local $_;
        while (<$file>) {
            next unless /^\d/;
            foreach my $w ($self->_parse_line($_, 1)) {
                my $word = $w->{word};
                $hash->{"$key:$word"}++;
            }
        }
    }
    return $hash;
}

sub write {
    my $self = shift;
    my ($ss, $fh) = @_;
    my $ret = "";

    my $off;
    eval { $off = $fh->getpos; 1; } or $off = $ss->{file_offset};
    $off ||= 0;
    $ret .= sprintf "%08d ", $off;
    $ret .= sprintf "%02d ", $ss->{lexfile};
    $ret .= $ss->pos . " ";
    $ret .= sprintf "%02x ", scalar(@{$ss->{words}});
    my $pcount = 0;
    foreach my $w (@{$ss->{words}}) {
        my $part = $w->word;
        $part =~ tr/ /_/;
        $part .= $w->syntax ? "(".$w->syntax.") " : " ";
        $part .= sprintf "%1x ", $w->lexsense;
        $ret .= $part;
        $pcount += scalar @{[$w->pointers]};
    }
    $pcount += scalar @{[$ss->pointers]};
    $ret .= sprintf "%03d ", $pcount;
    foreach my $p (map { ($_->pointers); } $ss, @{$ss->{words}}) {
        my ($to, $from) = map { $p->{$_}{id}; } qw/to from/;
        my ($pos, $offs, $tar) = split /:/, $to;
        (undef, undef, my $src) = split /:/, $from;
        $tar ||= 0;
        $src ||= 0;
        $ret .= sprintf "%s %08d %s %02x%02x ", $p->type, $offs, $pos, $src, $tar;
    }
    $pcount = 0;
    if (grep { $_->frames and scalar @{[$_->frames]}; } $ss, @{$ss->{words}}) {
        $pcount = scalar @{[$ss->frames]};
        foreach (0..scalar(@{$ss->{words}})-1) {
            $pcount+= scalar @{[ $ss->{words}[$_]->frames ]};
        }
        $ret .= sprintf "%02d ", $pcount;
        foreach ($ss->frames) {
            $ret .= sprintf "+ %02d %02x ", $_, 0;
        }
        for my $i (0..scalar(@{$ss->{words}})-1) {
            foreach my $f ($ss->{words}[$i]->frames) {
                $ret .= sprintf "+ %02d %02x ", $f, $i+1;
            }
        }
    }
    $ret .= "| ". $ss->gloss("original");
    $ret .= "  \n";
    return $ret;
}

sub _parse_line_re {
    my $self = shift;
	(local($_), my($just_words)) = (@_, 0);

    my $synset = { frames => [], pointers => {} };
    $synset->{unparsed} = $_;

	warn "Bad line: $_\n" and return unless s/^(\d{8})\s(\d\d)\s([nvars])\s([0-9a-f]{2})\s//i;
	@{$synset}{qw/file_offset lexfile pos/} = ($1, $2, $3);
	my $i = hex $4;
	$synset->{pos} = 'a' if $self->{conflate35} and $synset->{pos} eq 's';
    $synset->{id}  = "$synset->{pos}\:$synset->{file_offset}";
    while ($i-- > 0) {
        my $word = { frames => [], pointers => {} };
		warn "Invalid word\n" and return unless s/^(\S+)\s([0-9a-f])\s//i;
        (@{$word}{qw/word lexsense/}) = ($1, hex $2);
        $word->{word} =~ tr/_/ / unless $self->{retain_underscore};
        $word->{syntax} = ($word->{word} =~ s/\((a|p|ip)\)$//) ? $1 : '';
        $word->{pos} = $synset->{pos};
        $word->{id} = "$synset->{pos}\:$synset->{file_offset}\:" .
            ((defined $synset->{words}) ? @{$synset->{words}} + 1 : 1);
        push @{$synset->{words}}, WordNet::Word->new(%$word,
                                                     WN => $self->{WN});
    }
    return @{$synset->{words}} if $just_words;
	warn "Bad pointer count\n" and return unless s/^(\d{3})\s//;
    $i = int $1;
    while ($i-- > 0) {
		warn "Bad pointer: $_\n" and return unless s/^(\S+)\s(\d{8})\s([nvars])\s([0-9a-f]{2})([0-9a-f]{2})\s//i;
        my ($ptr, $off, $ptr_pos, $src, $tar) = ($1, $2, $3, $4, $5);
		$_ = hex $_ for $src, $tar;
        my $pointer = {
            type => $ptr,
            from => {
                class => "WordNet\::" . (($src) ? "Word" : "Synset"),
                pos   => (($src)
                          ? $synset->{words}[$src-1]->{pos}
                          : $synset->{pos} ),
                id    => "$synset->{pos}\:$synset->{file_offset}" .
                    (($src) ? ":" . int $src : "")
                },
            to   => {
                class => "WordNet\::" . (($tar) ? "Word":"Synset"),
                pos   => $ptr_pos,
                id    => "${ptr_pos}:$off" . (($tar) ? ":" . int $tar : "")
                }
        };
        $pointer->{toclass} = $pointer->{to}{class};
        $pointer->{id}      = "$pointer->{from}{id}:$pointer->{to}{id}";
        $pointer->{pos}     = "$pointer->{from}{pos}:$pointer->{to}{pos}";

        my $p = WordNet::Pointer->new(%$pointer, WN => $self->{WN});
        if ($src) {
            push @{$synset->{words}[$src-1]{pointers}{$p->type}}, $p;
        } else {
            push @{$synset->{pointers}{$p->type}}, $p;
        }
    }
	if (s/^(\d{2})\s//) {
        $i = int $1;
        while ($i-- > 0) {
			warn "Bad frame\n" and return unless s/^\+\s(\d\d)\s//;
            my $frame = int $1;
			warn "Bad frame target\n" and return unless s/^([0-9a-f]{2})\s//i;
            my $word = hex $1;
            if ($word) {
                push @{$synset->{words}[$word-1]{frames}}, $frame;
            } else {
                push @{$synset->{frames}}, $frame;
            }
        }
    }
	warn "Bad gloss\n" and return unless s/^\|\s//;
    my $gloss = $_;
    $gloss =~ s/\s+$//;
    $synset->{gloss}{original} = $gloss;
    my $me = WordNet::Synset->new(%$synset, WN => $self->{WN});
    $_->{SS} = $me for $me->Words;
    weaken($_->{SS}) for $me->Words;
    return $me;
}

sub _parse_line {
    my $self = shift;
    my ($line, $just_words) = (@_, 0);

    my $synset = { frames => [], pointers => {} };
    $synset->{unparsed} = $line;

    my @dat = split ' ', $synset->{unparsed};
    return undef unless @dat;
    $synset->{file_offset} = shift @dat;
    $synset->{lexfile}     = shift @dat;
    my $possible_pos       = shift @dat;
    $synset->{pos}         = (($possible_pos eq 's' and $self->{conflate35})
                              ? 'a'
                              : $possible_pos);
    $synset->{id}          = "$synset->{pos}\:$synset->{file_offset}";
    warn "invalid line (no word count): $line\n" and return undef
        unless @dat;
    warn "invalid line (word count: $dat[0]): $line\n" and return undef
        unless $dat[0] =~ /^[0-9a-f]{2}$/i;
    my $i = hex shift @dat;
    while ($i-- > 0) {
        my $word = { frames => [], pointers => {} };
        warn "invalid line (missing word): $line\n" and return undef
            unless @dat;
        $word->{word} = shift(@dat);
        $word->{word} =~ tr/_/ / unless $self->{retain_underscore};
        $word->{syntax} = ($word->{word} =~ s/\((a|p|ip)\)$//) ? $1 : '';
        warn "invalid line (missing lexical sense number): $line\n" and
            return undef unless @dat;
        warn "invalid line (lexical sense number: $dat[0]): $line\n" and
            return undef unless $dat[0] =~ /^[0-9a-f]$/i;
        $word->{lexsense} = hex shift @dat;
        $word->{pos} = $synset->{pos};
        $word->{id} = "$synset->{pos}\:$synset->{file_offset}\:" .
            ((defined $synset->{words}) ? @{$synset->{words}} + 1 : 1);
        push @{$synset->{words}}, WordNet::Word->new(%$word,
                                                     WN => $self->{WN});
    }
    return @{$synset->{words}} if $just_words;
    warn "invalid line (no pointer count): $line\n" and return undef
        unless @dat;
    warn "invalid line (pointer count: $dat[0]): $line\n" and return undef
        unless $dat[0] =~ /^\d{3}$/;
    $i = int shift @dat;
    while ($i-- > 0) {
        warn "invalid line (bad pointer): $line\n" and return undef
            unless @dat >= 4;
        my ($ptr, $off, $ptr_pos, $src_tar) = splice @dat, 0, 4;
        warn "invalid line (bad source/target: $src_tar): $line\n" and
            return undef unless $src_tar =~ /^[0-9a-f]{4}$/i;
        my $src = hex substr $src_tar, 0, 2;
        my $tar = hex substr $src_tar, 2, 2;
        my $pointer = {
            type => $ptr,
            from => {
                class => "WordNet\::" . (($src) ? "Word" : "Synset"),
                pos   => (($src)
                          ? $synset->{words}[$src-1]->{pos}
                          : $synset->{pos} ),
                id    => "$synset->{pos}\:$synset->{file_offset}" .
                    (($src) ? ":" . int $src : "")
                },
            to   => {
                class => "WordNet\::" . (($tar) ? "Word":"Synset"),
                pos   => $ptr_pos,
                id    => "${ptr_pos}:$off" . (($tar) ? ":" . int $tar : "")
                }
        };
        $pointer->{toclass} = $pointer->{to}{class};
        $pointer->{id}      = "$pointer->{from}{id}:$pointer->{to}{id}";
        $pointer->{pos}     = "$pointer->{from}{pos}:$pointer->{to}{pos}";

        my $p = WordNet::Pointer->new(%$pointer, WN => $self->{WN});
        if (int $src) {
            push @{$synset->{words}[$src-1]{pointers}{$p->type}}, $p;
        } else {
            push @{$synset->{pointers}{$p->type}}, $p;
        }
    }
    if (@dat and ($dat[0] ne '|')) {
        warn "invalid line (frame count: $dat[0]): $line\n" and return undef
            unless $dat[0] =~ /^\d{2}$/;
        $i = int shift @dat;
        while ($i-- > 0) {
            warn "invalid line ('+' expected, found '${\do{@dat?$dat[0]:''}}'): $line\n" and return undef unless @dat and $dat[0] eq '+';
            shift @dat;
            warn "invalid line (frame not found): $line\n" and return undef
                unless @dat;
            warn "invalid line (frame: $dat[0]): $line\n" and return undef
                unless $dat[0] =~ /^\d{2}$/;
            my $frame = int shift @dat;
            warn "invalid line (frame target not found): $line\n" and
                return undef unless @dat;
            warn "invalid line (frame target: $dat[0]): $line\n" and
                return undef unless $dat[0] =~ /^[0-9a-f]{2}$/i;
            my $word = hex shift @dat;
            if (int $word) {
                push @{$synset->{words}[$word-1]{frames}}, $frame;
            } else {
                push @{$synset->{frames}}, $frame;
            }
        }
    }
    warn "invalid line (no gloss): $line\n" and return undef unless @dat;
    warn "invalid line ('|' expected, found '$dat[0]'): $line\n"
        and return undef unless $dat[0] eq '|';
    shift @dat;
    my $gloss = join ' ', @dat;
    $gloss =~ s/\s+$//;
    $synset->{gloss}{original} = $gloss;
    my $me = WordNet::Synset->new(%$synset, WN => $self->{WN});
    $_->{SS} = $me for $me->Words;
    weaken($_->{SS}) for $me->Words;
    return $me;
}

sub parent_ss {
    my $self = shift;
    my $w = shift;
    my $id = $w->id;
    $id =~ s/:\d+$//;
    $self->read($id);
}

1;
__END__
# Below is stub documentation for your module. You'd better edit it!

=head1 NAME

WordNet::IO::Disk - Perl extension for blah blah blah

=head1 SYNOPSIS

  use WordNet::IO::Disk;
  blah blah blah

=head1 ABSTRACT

  This should be the abstract for WordNet::IO::Disk.
  The abstract is used when making PPD (Perl Package Description) files.
  If you don't want an ABSTRACT you should also edit Makefile.PL to
  remove the ABSTRACT_FROM option.

=head1 DESCRIPTION

Stub documentation for WordNet::IO::Disk, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited.

Blah blah blah.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Ben Haskell, E<lt>bhaskell@localdomainE<gt>

=head1 COPYRIGHT AND LICENSE

Copyright 2003 by Ben Haskell

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
