package WordNet::Morphology;

use 5.006;
use strict;
use warnings;

use WordNet qw/:pos/;
use WordNet::Filenames;
use Text::Balanced qw/extract_bracketed/;

require Exporter;

our $debug = 0;

our @ISA = qw(Exporter);

my @export = qw/&moan &groan &preserve_case/;
my @wn_vars = qw/@pos @pos_extended %pos_medium %pos_long/;
our %EXPORT_TAGS = ( 'all'       => [ @export, @wn_vars ],
                     'functions' => [ @export ],
                     'pos'       => [ qw/@pos @pos_extended
                                      %pos_medium %pos_long/ ] );
our (@EXPORT_OK)   = (@export, @wn_vars);
our @EXPORT        = ();
our $VERSION       = '0.01';

my %sub_parts = ( n => [ qw/n ns ng nsg/ ],
                  v => [ qw/v vs ved vpp ving/ ],
                  a => [ qw/a aer aest/ ],
                  r => [ qw/r rer rest/ ] );
foreach my $pos (@pos) {
    foreach my $part (@{$sub_parts{$pos}}) {
        next if $part eq $pos;
        $sub_parts{$part} = [ $part ];
    }
}
my @part_order = map { @{$sub_parts{$_}} } @pos;
my %part_order = map { ($part_order[$_], $_ + 1) } (0..$#part_order);
my %super_part;
foreach my $pos (@pos) {
    %super_part = (%super_part, map { ($_, $pos) } (@{$sub_parts{$pos}}));
}

our @defaults = ( morphdir       => 'wnsearchdir',
                  irrdir         => 'morphdir',
                  use_wn         => 1,
                  case_sensitive => 0,
                  inf_file       => 'patterns.inf',
                  assume_ending  => 1 );

sub new {
    my $class = shift;
    my $ref = { @defaults, @_ };
    $debug = {@_}->{debug} if {@_}->{debug};

    my $filenames = WordNet::Filenames->new(@defaults, @_);

	for (map([$_, $pos_medium{$_}], @pos),
	     map([$_, $_],              qw/ignore various/),
	    ) {
		my ($key, $name) = @$_;
		$name .= '.irr';
		$ref->{filenames}{irr}{$key} = $filenames->catfile('irr', $name);
	}
    $ref->{filenames}{inflect} = $filenames->catfile('morph',$ref->{inf_file});
    $ref->{filenames}{generate}  = $filenames;

    $ref->{wordnet} = WordNet->new(%$ref)
        if $ref->{use_wn} and not $ref->{wordnet};

    my $self = bless $ref, $class;
    $self->initialize or die 'Error initializing morphology routines';
    $self;
}

sub initialize {
    my $self = shift;
    $self->_load_files() or die 'Error opening morphology files';
    $self->_init_irr()   or die 'Error loading irregulars files';
    $self->_init_inf()   or die 'Error loading inflection files';
    return 1;
}

my $sub_re = qr/^([A-Z])$/;
my $direction_re = qr/^(-(?:-|>))$/;
my $name_re = qr/((?:\w|-)+)/;
my $label_re = qr/^$name_re:$/;
my $pos_re = join '|', map { reverse @{$sub_parts{$_}} } @pos;
$pos_re = qr/^($pos_re)$/;

sub _init_irr {
    my $self = shift;
    my %formats = ( v => { 3 => [ qw/v ved vpp/ ],
                           4 => [ qw/v ved vpp ving/ ],
                           7 => [ qw/v vs vs vs ved ved vpp/ ] },
                    n => { 2 => [ qw/n ns/ ] },
                    a => { 3 => [ qw/a aer aest/ ] },
                    r => { 3 => [ qw/r rer rest/ ] },
                    ignore => { 1 => [ qw/ignore/ ] } );
    foreach my $pos (@pos, 'ignore') {
        my $file = $self->{files}{irr}{$pos};
        local $_;
        while (<$file>) {
            next if (/^\#/ or /^\s*$/);
            chomp;
            s/\s*\#.*$//; # remove comments
            my ($line, $prefixes) = split /\s+:\s+/;
            my @split = split /\s+/, $line;
            tr/_/ / for @split;
            $_ = [ split /\// ] for @split;
            my @prefixes;
            @prefixes = split /\s/, $prefixes if $prefixes;
            tr/_/ / for @prefixes;

            unless (defined $formats{$pos}{scalar @split}) {
                print STDERR "Invalid irregulars file ($pos) line: " .
                    join(' ', @split) . "\n";
                next;
            }
            if ($pos ne 'ignore') {
                foreach my $pre ('', @prefixes) {
                    my @forms = map { [ map { $pre.$_ } @$_ ] } @split;
                    my @parts = @{$formats{$pos}{scalar @split}};
                    @parts = map {
                        [ ($parts[$_]) x scalar @{$split[$_]} ]
                        } 0..$#split;
                    my @bases = @{shift @forms};
                    my @base_parts = @{shift @parts};
                    for my $i (0..$#forms) {
                        my @infl = @{$forms[$i]};
                        my @parts = @{$parts[$i]};
                        for my $b (0..$#bases) {
                            for my $j (0..$#infl) {
                                push @{$self->{irr}{moan}{$parts[$j]}{$infl[$j]}},
                                $bases[$b], $base_parts[$b];
                                push @{$self->{irr}{groan}{$base_parts[$b]}{$bases[$b]}},
                                $infl[$j], $parts[$j];
                            }
                        }
                    }
                }
            } else {
                my @w = @{shift @split};
                for my $p (@pos) {
                    for my $w (@w) {
                        for my $pre ('', @prefixes) {
                            my $W = $pre.$w;
                            push @{$self->{irr}{ignore}{moan}{$_}{$W}}, $W, $p
                                for @{$sub_parts{$p}};
                            push @{$self->{irr}{ignore}{groan}{$p}{$W}}, $W, $_
                                for @{$sub_parts{$p}};
                        }
                    }
                }
            }
        }
    }
    my $file = $self->{files}{irr}{various};
    local $_;
    while (<$file>) {
        next if /^\#/;
        chomp;
        my @pieces = split /\|/;
        warn "Bad line in various: $_\n" and next unless @pieces == 4;
        warn "Bad part of speech ('$pieces[1]' or '$pieces[3]') in various in line: $_\n" and next if grep { !/$pos_re/ } @pieces[1,3];
        my ($bw, $bp, $iw, $ip) = @pieces;
        push @{$self->{irr}{moan}{$ip}{$iw}}, $bw, $bp;
        push @{$self->{irr}{groan}{$bp}{$bw}}, $iw, $ip;
    }
    return 1;
}

sub _parts_of_speech {
    local $_ = shift;
    my $line = shift;
    my @ret = (/$pos_re/);
    return @ret if @ret;
    return () unless s/^\{(.*)\}$/$1/;
    @ret = split ',';
    return () if grep { !/$pos_re/ } @ret;
    @ret;
}

sub _patterns {
    local $_ = shift;
    my $line = shift;
    my @ret = ();
    my ($prefix, $postfix) = ('', '');
    ($prefix, $postfix) = ($1, $3) if s/(.*?)\{([^\}]+)\}(.*)/$2/;
    foreach my $part (split ',') {
        push @ret, "$prefix$part$postfix";
    }
    @ret;
}

sub _functions {
    my $self = shift;
    local $_ = shift;
    my $line = shift;
    return () if /$direction_re/;
    my @ret = split ',';
    foreach (@ret) {
        die "Invalid function name $_:\nline $.: $line" unless /^$name_re$/;
        die "Undefined function $_ used in line:\n$.: $line" unless defined $self->{functions}{$_};
    }
    @ret;
}

my %command = ( c => \&add_collocation_pattern,
                s => \&add_substitution,
                o => \&set_option,
                f => \&add_function,
                e => sub { shift; eval "@_"; } );

sub add_collocation_pattern {
    my $self = shift;
    my @parts = _parts_of_speech(shift);
    my $pattern = shift;
    return 0 unless @parts and $pattern;
	my @arr = split //, $pattern;
	my @which = map 2*$_, grep $arr[$_] eq 'X', 0..$#arr;
	push @{$self->{collocations}{$_}{2*@arr-1}}, [ @which ] for @parts;
    return 1;
}

sub add_substitution {
    my $self = shift;
    my ($char, $replace) = @_;
    $replace = qr/$replace/i;
    return 0 if $char !~ /$sub_re/;
    $self->{substitutions}{$char} = $replace;
    return 1;
}

sub set_option {
    my $self = shift;
    my ($option, $value) = @_;
    return 0 if $option !~ /^$name_re$/;
    return 0 if not defined $self->{$option};
    $value = $self->{$option} ? 0 : 1 if not defined $value;
    $self->{$option} = $value;
    return 1;
}

sub add_function {
    my $self = shift;
    my $name = shift;
    my $subr = "@_";
    return 0 unless $name =~ /^$name_re$/;
    return 0 if defined $self->{$name};
    $self->{functions}{$name} = eval $subr
        or die "Error initializing subroutine:\n$.: $subr";
    return 1;
}

sub _init_inf {
    my $self = shift;
    my $fh = $self->{files}{inflect};
    local $_;
    while (<$fh>) {
        next if /^\#/ or /^\s*$/;
        my $line = $_;
        my $label = "line_$.";
        my @parts = split /\s+/;

        my $part = shift @parts;
        if ($part =~ /$label_re/) {
            $label = $1;
            $part = shift @parts;
        }

        my @start_pos = _parts_of_speech($part);
        if (!@start_pos) {
            defined $command{$part}
                or die "Bad command: $line";
            $command{$part}->($self, @parts)
                or die "Error with line: $line";
            next;
        }

        $part = shift @parts;
        my @start_patterns = _patterns($part, $line);
        @start_patterns = map { $_.'$' } @start_patterns
            if $self->{assume_ending};

        $part = shift @parts;
        my @start_functions = ();
        unless ($part =~ $direction_re) {
            @start_functions = $self->_functions($part, $line);
            $part = shift @parts;
        }
        my $two_way = ($part eq '--') ? 1 : 0;

        $part = shift @parts;
        my @end_pos = _parts_of_speech($part, $line);

        $part = shift @parts;
        my @end_patterns = _patterns($part, $line);
        @end_patterns = map { $_.'$' } @end_patterns
            if $self->{assume_ending};
        
        my @end_functions = ();
        if (@parts) {
            $part = shift @parts;
            @end_functions = $self->_functions($part, $line);
        }

        my $max = 0;
        foreach ($#start_pos, $#start_patterns, $#end_pos, $#end_patterns) {
            if ($_ and $max and $_ != $max) {
                die 'Mismatched number of patterns and parts of speech: '.
                    ($_+1).' vs. '.($max+1)."\n$line";
            }
            $max = $_ if $_;
        }
        my ($spos, $spat, $epos, $epat, $new_label);
        for (1..($max+1)) {
            $spos = shift @start_pos      || $spos;
            $spat = shift @start_patterns || $spat;
            $epos = shift @end_pos        || $epos;
            $epat = shift @end_patterns   || $epat;
            $new_label = "$label ($spos -> $epos)";

            $self->add_rule('groan', $line, $new_label.' groan',
                            $spos, $spat, \@start_functions,
                            $epos, $epat, \@end_functions);
            next unless $two_way;
            my ($new_spat, $new_epat) = $self->_reverse($spat, $epat);
            $self->add_rule('moan', $line, $new_label.' moan',
                            $epos, $new_spat, \@end_functions,
                            $spos, $new_epat, \@start_functions);
        }
    }
    return 1;
}

sub do_subs {
    my $self = shift;
    my $pattern = shift;
    foreach my $sub (keys %{$self->{substitutions}}) {
        $pattern =~ s/$sub/$self->{substitutions}{$sub}/g;
    }
    $pattern;
}

sub add_rule {
    my $self = shift;
    my ($which,
        $original, $label,
        $spos, $spat, $sf,
        $epos, $epat, $ef) = @_;
    $epat =~ s/\Q()\E/\$1/; # () --> $1
    $epat =~ s/\$$//;       # remove trailing '$'
    $epat =~ s/\\1/\$1/;    # \1 --> $1
    my $ssub = $self->do_subs($spat);
    $ssub = qr/$ssub/;
    my $rule = { label => $label,
                 pos => $spos,
                 pattern => $spat,
                 subbed => $ssub,
                 functions => [(@$sf)],
                 to => { pos => $epos,
                         pattern => $epat,
                         functions => [(@$ef)] },
                 original => $original
                 };
    push @{$self->{rules}{$which}{$spos}}, $rule;
}

sub _reverse {
    my $self = shift;
    my ($spat, $epat) = @_;
    my $ignore = qr/([^\(]|\(\?)*/; # ignore leading (?...) clauses
    my ($scapt, $spost, $spre) = extract_bracketed($spat, '()', $ignore);
    my ($ecapt, $epost, $epre) = extract_bracketed($epat, '()', $ignore);
    $spat = join '', $spre, $scapt, $epost;
    $epat = join '', $epre, $ecapt, $spost;
    return ($spat, $epat);
}

sub _load_files {
    my $self = shift;
    foreach my $file (keys %{$self->{filenames}{irr}}) {
        open $self->{files}{irr}{$file}, "<$self->{filenames}{irr}{$file}"
            or die "Couldn't open $self->{filenames}{irr}{$file}: $! $@";
    }
    open $self->{files}{inflect}, "<$self->{filenames}{inflect}"
        or die "Couldn't open $self->{filenames}{inflect}: $! $@";
    return 1;
}

sub preserve_case {
    my ($old, $new) = @_;
    my $mask = uc $old ^ $old;
    $mask .= substr($mask, -1) x (length($new) - length($old));
    $mask = substr($mask,0,length($new));
    uc $new | $mask;
}

sub _table_lookup {
    my $self = shift;
    my ($table, $word, $pos) = @_;
    my $lc = lc $word;
    my @ret = $table->{$pos}{$lc} ? @{$table->{$pos}{$lc}} : ();
    push @ret, ($table->{$pos}{$word} ? @{$table->{$pos}{$word}} : ())
        if $self->{case_sensitive};
    $ret[$_] = preserve_case($word, $ret[$_]) for grep { not $_ % 2 } 0..$#ret;
    @ret;
}    
sub exceptions {
    my $self = shift;
    my ($which, $word, $pos) = @_;
    my $table = $self->{irr}{$which};
    $self->_table_lookup($table, $word, $pos);
}
sub ignores {
    my $self = shift;
    my ($which, $word, $pos) = @_;
    my $table = $self->{irr}{ignore}{$which};
    $self->_table_lookup($table, $word, $pos);
}

sub _analyze {
    my $self = shift;
    my ($which, $word, $pos) = @_;
    my @ret = ();

    if (my @a = $self->ignores($which, $word, $pos)) {
        print STDERR "ignored: $word $pos -> (@a)\n" if $debug;
        while (my ($w, $p) = splice @a, 0, 2) {
            push @ret, $w, $p, 'ignores';
        }
        return @ret;
    }

    my (@piece) = split /(\s|-|_)/, $word;
    return @ret unless @piece;

    my @collocation_patterns = ([$#piece],
                                defined($self->{collocations}{$super_part{$pos}}{scalar @piece})
                                ? @{$self->{collocations}{$super_part{$pos}}{scalar @piece}}
                                : ());
    foreach my $coll ($#piece ? @collocation_patterns : ()) {
        my %new_piece;
        for (grep { $_ % 2 ? 0 : 1 } 0..$#piece) {
            my @an = $self->_analyze($which, $piece[$_], $pos);
            while ((my ($neww, $newp), undef) = splice @an, 0, 3) {
                push @{$new_piece{$newp}[$_]}, $neww;
            }
        }
        for my $k (keys %new_piece) {
            for my $n (0..$#piece) {
                if ($new_piece{$k}[$n]) {
                    $new_piece{$k}[$n] =
                        [ keys %{{(map { $_ => 1 } @{$new_piece{$k}[$n]})}} ];
                } else {
                    $new_piece{$k}[$n] = [ $piece[$n] ];
                }
            }
        }
        for my $p (keys %new_piece) {
            my @w = @{$new_piece{$p}[0]};
            for my $n (1..$#{$new_piece{$p}}) {
                my @add;
                for my $add (@{$new_piece{$p}[$n]}) {
                    push @add, map { $_.$add } @w;
                }
                @w = @add;
            }
            push @ret, $_, $p, 'done' for @w;
        }
    }
    my %exceptional;
    if (my @a = $self->exceptions($which, $word, $pos)) {
        print STDERR "exceptional: $word $pos -> (@a)\n" if $debug;
        while (my ($w, $p) = splice @a, 0, 2) {
            push @ret, $w, $p, 'exceptions';
            $exceptional{$p}++;
        }
    }
    foreach my $rule (@{$self->{rules}{$which}{$pos}}) {
        my ($rpat, $repl) = ($rule->{subbed}, $rule->{to}{pattern});
        foreach my $coll (@collocation_patterns) {
            my @coll = @$coll;
            if (grep { $piece[$_] !~ $rpat } @coll) {
                print STDERR "REJECT <$rule->{label}>: !~ $rule->{pattern}\n" if $debug > 3;
                next;
            }
            my @new_piece = @piece;
            for (@coll) {
                my @sub = ($piece[$_] =~ /$rpat/);
                my $rep = $repl;
                $rep =~ s/\$$_/$sub[$_-1]/g for 1..@sub;
                $new_piece[$_] =~ s/$rpat/$rep/;
            }
            my $new_pos = $rule->{to}{pos};
            my $new_word = join '', @new_piece;
            my @parameters = ($word, $pos, $new_word, $new_pos);
            if (my @a = grep(!$self->{functions}{$_}->(@parameters),
                             @{$rule->{functions}})) {
                print STDERR "FUNCTION: @a\n" if $debug > 2;
                next;
            }
            @parameters = @parameters[2,3,0,1];
            if (my @a = grep(!$self->{functions}{$_}->(@parameters),
                             @{$rule->{to}{functions}})) {
                print STDERR "TO_FUNCTION: @a\n" if $debug > 2;
                next;
            }
            print STDERR "SUCCESS: <$rule->{label}>: $word $pos -> $new_word $new_pos\n" if $debug;
            print STDERR "ORIGINAL: $rule->{original}\n" if $debug > 2;
            next if $exceptional{$new_pos};
            print STDERR "NOT_EXCEPTIONAL: $new_pos\n" if $debug > 1;
            push @ret, $new_word, $new_pos, "rule_${pos}_${new_pos}";
        }
    }
    @ret;
}

sub _de_dup_and_sort {
    my $ret = shift;
    my (@arr, %dups);
    for (my $i=0; $i<scalar @$ret; $i+=2) {
        my ($w, $p) = @$ret[$i,$i+1];
        next if $dups{$w}{$p}++;
        next unless $part_order{$p};
        push @arr, [ $w, $p, $part_order{$p} ];
    }
    @$ret = map((@{$_}[0,1]),
                sort { $a->[2] <=> $b->[2] or $a->[0] cmp $b->[0] } @arr);
}

sub is_word {
    my $self = shift;
    $self->{use_wn} ? $self->{wordnet}->is_word(@_) : 1;
}

sub moan {
    my $self = shift;
    my ($word, @parts) = (@_);
    @parts = @pos if not @parts;
    @parts = map { @{$sub_parts{$_}} } @parts;
    my @ret = map { ( ($super_part{$_} eq $_ ? ($word, $_, 'self') : ()),
                      $self->_analyze('moan', $word, $_)) } @parts;
    print STDERR "Back from _analyze: (@ret)\n" if $debug > 1;
    my @newret;

    while (my ($w, $p, $status) = splice @ret, 0, 3) {
        if ($status =~ /^rule_([a-z]+)_([a-z]+)/) {
            my ($opos, $npos) = ($1, $2);
            unless ($opos =~ /ns?g|vs|ving/) {
                my @bad = $self->exceptions('groan', $w, $p);
                my %bad;
                while (my ($bw, $bp) = splice @bad, 0, 2) {
                    $bad{$bw}{$bp}++;
                }
                unless ($bad{$w}{$p}) {
                    for my $k (keys %bad) {
                        for my $j (keys %{$bad{$k}}) {
                            push @bad, $k, $j;
                        }
                    }
                }
                $debug and warn "bad: $w $p -> (@bad)\n" if @bad;
                next if @bad;
            }
            if ($npos eq 'n') {
                my @exc = map {
                    $self->exceptions('moan', $w, $_);
                } @{$sub_parts{$super_part{$p}}};
                $debug and warn "double: $w $p -> (@exc)\n" if @exc;
                push @newret, @exc;
            }
        }
        push @newret, $w, $p;
    }
    @ret = @newret;

    _de_dup_and_sort(\@ret);
    return @ret unless $self->{use_wn};
    my @actual_ret = ();
    while (my ($w, $p) = splice @ret, 0, 2) {
        if ($p eq $super_part{$p} and $self->{wordnet}->is_word($w, $p)) {
            push @actual_ret, $w, $p;
            print STDERR "IS__WORD: $w $p\n" if $debug > 1;
        } else {
            print STDERR "NOT_WORD: $w $p\n" if $debug > 1;
        }
    }

    # RAT/RATE HACK - START
    my %verbs;
    for (0..(@actual_ret/2 - 1)) {
        $verbs{$actual_ret[$_*2]}++ if $actual_ret[$_*2+1] eq 'v';
    }
    my %rat_word;
    for (keys %verbs) {
        # next u.l. /  (^ or C) V C(not wxy) $ /
        next unless /(?:^|[^aeiou])[aeiou][^aeiouwxy]$/;
        $rat_word{$_}++ if $verbs{$_.'e'};
    }
    while (my ($w, $p) = splice @actual_ret, 0, 2) {
        if ($p ne 'v' or not $rat_word{$w}) {
            print STDERR "NOT__RAT: $w\n" if $debug > 1;
            push @ret, $w, $p;
        } else {
            print STDERR "RAT_WORD: $w\n" if $debug > 1;
        }
    }
    # RAT/RATE HACK - END
    @ret;
}

sub groan {
    my $self = shift;
    my ($word, @parts) = (@_);
    @parts = @pos unless @parts;
    @parts = grep { $self->is_word($word, $_) } @parts;
    my @ret = map { ( ($super_part{$_} eq $_ ? ($word, $_, 'self') : ()),
                      $self->_analyze('groan', $word, $_)) } @parts;
    print STDERR "Back from _analyze: (@ret)\n" if $debug > 1;
    my @newret;

    while (my ($w, $p, $status) = splice @ret, 0, 3) {
        if ($status =~ /^rule_([a-z]+)_([a-z]+)/) {
            my ($opos, $npos) = ($1, $2);
            unless ($npos =~ /ns?g|vs|ving/) {
                my @bad = $self->exceptions('moan', $w, $p);
                my %bad;
                while (my ($bw, $bp) = splice @bad, 0, 2) {
                    $bad{$bw}{$bp}++;
                }
                unless ($bad{$w}{$p}) {
                    for my $k (keys %bad) {
                        for my $j (keys %{$bad{$k}}) {
                            push @bad, $k, $j;
                        }
                    }
                }
                $debug and warn "bad: $w $p -> (@bad)\n" if @bad;
                next if @bad;
            }
#             if ($npos eq 'n') {
#                 my @exc = map {
#                     $self->exceptions('groan', $w, $_);
#                 } @{$sub_parts{$super_part{$p}}};
#                 $debug and warn "double: $w $p -> (@exc)\n" if @exc;
#                 push @newret, @exc;
#             }
        }
        push @newret, $w, $p;
    }
    @ret = @newret;

    _de_dup_and_sort(\@ret);

    my @actual_ret = @ret; @ret = (); # HACK HACK
    # RATED/RATTED HACK - START
    # [bdfgklmnprtsv]   # Double-able consonants
    my $endings = qr/(?=e(?:st|[rsd])|ing$ )/x;
    my $short_is_rat = ($word =~ /e$/i) ? 0 : 1;
    my %words;
    for (0..$#actual_ret) {
        next if $_ % 2;
        my ($form, $part) = @actual_ret[$_,$_+1];
        (my $short = $form) =~ s/([bdfgklmnprtsv])\1$endings/$1/;
        $short =~ s/e$endings//;
        push @{$words{$part}{$short}}, $form;
    }
    my %rat_word;
    for my $part (keys %words) {
        for my $short (keys %{$words{$part}}) {
            my $choices = $words{$part}{$short};
            next unless @$choices > 1;
            my @out = sort { length $a <=> length $b } @$choices;
            my $rat = $short_is_rat ? $out[0] : $out[-1];
            $rat_word{$rat}{$part}++;
        }
    }
#    use Data::Dumper; print Dumper \%words, \%rat_word;
    while (my ($w, $p) = splice @actual_ret, 0, 2) {
        if (not $rat_word{$w}{$p}) {
            print STDERR "NOT__RATED: $w\n" if $debug > 1;
            push @ret, $w, $p;
        } else {
            print STDERR "RATED_WORD: $w\n" if $debug > 1;
        }
    }
    # RATED/RATTED HACK - END
    @ret;
}

1;

__END__
# TODO - Edit Documentation

=head1 NAME

WordNet::Morphology - Perl extension for blah blah blah

=head1 SYNOPSIS

  use WordNet::Morphology;
  blah blah blah

=head1 ABSTRACT

  This should be the abstract for WordNet::Morphology.
  The abstract is used when making PPD (Perl Package Description) files.
  If you don't want an ABSTRACT you should also edit Makefile.PL to
  remove the ABSTRACT_FROM option.

=head1 DESCRIPTION

Stub documentation for WordNet::Morphology, created by h2xs. It looks like the
author of the extension was negligent enough to leave the stub
unedited. But looks can be deceiving...

Blah blah blah.

=head2 EXPORT

None by default.



=head1 SEE ALSO

Mention other useful documentation such as the documentation of
related modules or operating system documentation (such as man pages
in UNIX), or any relevant external documentation such as RFCs or
standards.

If you have a mailing list set up for your module, mention it here.

If you have a web site set up for your module, mention it here.

=head1 AUTHOR

Ben Haskell, E<lt>bhaskell@clarity.princeton.eduE<gt>

=head1 COPYRIGHT AND LICENSE

TODO -- Insert WordNet License

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself. 

=cut
