package node;

use 5.006;
use strict;
use warnings;
#use lib '/wordnet/wn/2k3/lib/perl';
use lib '.';
use WordNet qw/:vars/;
use WordNet::Morphology;
use CGI qw/:standard/;
use Scalar::Util qw/weaken/;

# "node" Class: (blessed hash ref)
#    -   OBJECT (querydata string/Word object/Synset object)
#    -   RELATEDTO (ref to array of Word objects)
#            (holds info for "Indirect via", "Opposed to", etc.)
#    -   SUPPRESSED (0,1)
#    -   PARENT (node ref)
#    -   CHILDREN (ref to array of node refs)
#    -   COUNT (index of node in history list)
#    -   MORPH ($node->{'MORPH'}->{part of speech} = '|'-separated list of the
#               strings returned from $morph->moan for that part of speech)
#            (used by getSCcount())
#    -   EXTRA (anything; used for abnormal stuff like sentence frames)

my $SYN_ID_LENGTH = 10;

#Extensions to WordNet database relations:
our %extRels = ('over', 'Overview', 'syno', 'synset',
               'sfrm', 'sentence frame');

my $wn = WordNet->new( nowarnings => 1 );
our $morph = WordNet::Morphology->new(wordnet => $wn,
                    morphdir => './WordNet/various',
                    case_sensitive => 0);

# #CONSTRUCTORS:
# sub init;   #takes class, searchTerm; returns empty head node w/ children
# sub new;    #takes class, object; returns new node
# #SETTING & RETRIEVING VALUES:
# sub getobject;  #takes node; returns object
# sub type;       #takes node; returns type (r,s,w)
# sub relatedto;  #takes node (& list); (sets &) returns array ref
# sub suppressed; #takes node (& value); (sets &) returns value
# sub parent;     #takes node (& parent node); (sets &) returns parent node
# sub children;   #takes node (& list of kid nodes); (sets &) returns kid nodes
# sub count;      #takes node; returns count
# sub pos;        #takes node; returns pos
# sub clip;       #takes node; removes & returns kids
# sub gloss;      #takes node; returns gloss for Word/Synset
# sub getSCcount; #takes node; returns SemCor project frequency count
# sub extra;      #takes node; returns extra
# #DOING SEARCHES:
# sub getSynset;  #takes node; returns Synset (for Word/Synset)
# sub getOffset;  #takes node; returns byte offset from id
# sub finalize;   #takes head node & -1; sets COUNT for all nodes
# sub geth;       #takes head node; returns new h parameter string
# sub expand;     #takes node, recSearch, max(max nodes to add); returns new max
# sub unsuppress; #takes node; unsuppresses children
# sub printdebug; #takes head node; prints out data for entire list
# #PRIVATE UTILITY FUNCTIONS: (these aren't in order)
# sub expandR;    #takes node, relation, max; used by expand for Show All search
# sub xSearch;    #takes node, #, max; used by expand for extended search types
# sub truncate;   #takes node; cuts off children & nephews
# sub doSearch;   #takes node; does search & returns child nodes
# sub addRel;     #takes relation & results array ref; adds rel if not there
# sub wwSynsetRelSearch; #takes rel, parent Synset, \@results;
#                        #adds indirect (word-word) relations for parent's Words


#CONSTRUCTORS:

sub init {
    my ($class, $searchTerm) = @_;
    my $self = {};
    bless $self, $class;
    my @res = $morph->moan($searchTerm);
    my %morph = ();
    my @results = ();
    for (my $i = 0; $i < @res; $i += 2) {
        push(@results, $wn->fetch($res[$i], $res[$i + 1]));
        push(@{$morph{$res[$i + 1]}}, $res[$i]);
    }
    $self->children(map { $_ = node->new($_); } @results);
    foreach (@{$self->children}) {
        $_->parent($self);
    }
    $self->{'OBJECT'} = 'over';
    foreach (keys %morph) {
        $morph{$_} = join('|', @{$morph{$_}});
    }
    $self->{'MORPH'} = \%morph;
    return $self;
}


sub new {
    bless {'OBJECT', $_[1]}, $_[0];
}


#SETTING & RETRIEVING VALUES:

sub getobject {
    $_[0]->{'OBJECT'};
}


sub type {
	my $type = ref $_[0]->{OBJECT};
	return 'w' if $type =~ /:Word$/;
	return 's' if $type =~ /:Synset$/;
	return 'r';
}


sub relatedto {
    my $self = shift;
    if (@_) { push @{$self->{'RELATEDTO'}}, @_ }
    return $self->{'RELATEDTO'} || [];
}


sub suppressed {
    if (defined($_[1])) { return ($_[0]->{'SUPPRESSED'} = $_[1]) }
    return $_[0]->{'SUPPRESSED'};
}


sub parent {
    my $self = shift;
    return $self->{PARENT} unless @_;
    $self->{PARENT} = shift;
    weaken $self->{PARENT};
    $self->{PARENT};
}


sub children {
    my $self = shift;
    if (@_) { push @{$self->{'CHILDREN'}}, @_ }
    return $self->{'CHILDREN'} || [];
}


sub count { $_[0]->{'COUNT'} || 0 }


sub pos {
    my $id = $_[0]->{'OBJECT'}->id;
    $id =~ m/^(.).*$/g;
    return $1;
}


sub clip {
    my @kids = @{$_[0]->{'CHILDREN'}};
    $_[0]->{'CHILDREN'} = [];
    return @kids;
}


sub gloss {
    if ($_[0]->type eq 's') {
        return $_[0]->getobject->gloss;
    } else {
        return $_[0]->getSynset->gloss;
    }
}


sub getSCcount {
    my $morphs = $_[0]->parent->{'MORPH'}{$_[0]->pos};
    if (!$morphs) { return 0 }
    my @Words = $_[0]->getobject->Words;
    for (my $i = 0; $i < @Words; $i++) {
        if ($Words[$i]->word =~ /^($morphs)$/i) { return $Words[$i]->count }
    }
    return 0;
}


sub extra { $_[0]->{'EXTRA'} || '' }


#DOING SEARCHES:

sub getSynset {
    my $id = $_[0]->getobject->id;
    my $pos = join("", @pos_extended);
    $id =~ m/^([$pos]:[0-9]{8})/g;
    return $wn->fetch_by_id($1);
}


sub getOffset {
    my $id = $_[0]->getobject->id;
    $id =~ m/^.:([0-9]{8})/;
    return $1;
}


sub finalize { # pass empty head node & -1 to start
    my ($self, $count) = @_;
    $self->{'COUNT'} = $count;
    foreach (@{$self->children}) {
        $count = $_->finalize($count + 1);
    }
    return $count;
}


sub geth {
    my $h = 0;
    my @kids = @{$_[0]->children};
    if (@kids) { $h = 1; }
    if ($_[0]->{'SUPPRESSED'}) { $h += 2; }
    if ($_[0]->{'COUNT'} == -1) { $h = ''; }
    foreach (@kids) {
        $h .= $_->geth;
    }
    return $h;
}


sub expand {
    my ($self, $recSearch, $max) = @_; # if $recSearch is 1, $max is defined
    my @kids = @{$self->children};
    if (!@kids) {
        @kids = doSearch($self);
        if (defined($max)) {
            $max -= @kids;
            if ($max < 0) {
                $self->truncate;
                return $max;
            }
        }
        $self->children(@kids);
        foreach (@kids) { $_->parent($self); }
    }
    if ($recSearch) {
        if ($recSearch == 1) {
            if ($self->type ne 'r') { return $max; }
            foreach (@kids) {
                $max = $_->expandR($self->getobject, $max);
                if ($max < 0) {
                    $_->truncate;
                    return $max;
                }
            }
        } else {
            $max = $self->xSearch($recSearch, $max);
        }
    }
    return $max;
}


sub expandR {
    my ($self, $rel, $max) = @_;  # max is always defined
    my @relkids = @{$self->children};
    if (@relkids) {
        my $relNode = undef;
        foreach (@relkids) {
            $_->suppressed(1);
            if ($rel eq $_->getobject) { $relNode = $_ }
            else { $_->{'CHILDREN'} = [] }
        }
        if (!$relNode) {
            $self->{'CHILDREN'} = [];
            return $max;
        }
        my @obkids = @{$relNode->children};
        if (!@obkids) {
            $max = $relNode->expand(0, $max);
            if ($max < 0) { return $max; }
            @obkids = @{$relNode->children};
        }
        foreach (@obkids) {
            $max = $_->expandR($rel, $max);
            if ($max < 0) { return $max; }
        }
    } else {
        @relkids = doSearch($self);
        $max -= @relkids;
        if ($max < 0) {
            $self->truncate;
            return $max;
        }
        my @relNode = grep { $rel eq $_->getobject } @relkids;
        if (!@relNode) { return $max; }
        $self->children(@relkids);
        foreach (@relkids) {
            $_->parent($self);
            $_->suppressed(1);
        }
        $max = $relNode[0]->expand(0, $max);
        if ($max < 0) { return $max; }
        my @obkids = @{$relNode[0]->children};
        foreach (@obkids) {
            $max = $_->expandR($rel, $max);
            if ($max < 0) { return $max; }
        }
    }
    return $max;
}


sub xSearch {
    my ($self, $type, $max) = @_;
    if ($type == 2) {
        if ($self->getobject eq 'hype') {
            my @hypes = @{$self->children};
            if (!@hypes) {
                $max = $self->expand(0, $max);
                if ($max < 0) { return $max }
                @hypes = @{$self->children};
            }
            foreach (@hypes) {
                $max = $_->expand(0, $max);
                if ($max < 0) { return $max }
                my @rels = @{$_->children};
                my $hypoRel = undef;
                foreach (@rels) {
                    $_->suppressed(1);
                    if ($_->getobject eq 'hypo') { $hypoRel = $_ }
                    else { $_->{'CHILDREN'} = [] }
                }
                $hypoRel->{'CHILDREN'} = [];
                $max = $hypoRel->expand(0, $max);
                if ($max < 0) { return $max }
            }
        }
    }
    return $max;
}


sub truncate { # takes node; cuts off children & nephews
    my $self = shift;
    $self->{'CHILDREN'} = [];
    my @siblings = @{$self->parent->children};
    foreach (@siblings) {
        $_->{'CHILDREN'} = [];
    }
}


sub doSearch { # return nodes, not objects
    my $self = shift;
    my $object = $self->getobject;
    my @results = ();
    my $type = $self->type;
    if ($type eq 'r') { #relation (add Synsets/Words)
        if ($object eq 'dummy') { return () }
        my $prev = $self->parent;
        my $prevob = $prev->getobject;
        if ($prev->type eq 's') { # synset's relation
            foreach ($prevob->pointers($object)) {
                push(@results, node->new($_->follow));
            }
            if ($object eq 'sfrm') {
                foreach ($prevob->framestext) {
                    my $frame = node->new('dummy');
                    $frame->{'EXTRA'} = $_;
                    push(@results, $frame);
                }
            }
            wwSynsetRelSearch($object, $prevob, \@results);
            # and check for indirect antonyms:
            if (@results == 0 && $object eq 'ants' && $prev->pos eq 'a') {
                foreach ($prevob->pointers('sim')) {
                    wwSynsetRelSearch($object, $_->follow, \@results);
                }
            }
        } else { # word's relation
            if (!exists($extRels{$object})) {
                foreach ($prevob->pointers($object)) {
                    push(@results, node->new($_->follow));
                }
            } elsif ($object eq 'over') {
                my @res = $morph->moan($prevob->word);
                my %morph = ();
                for (my $i = 0; $i < @res; $i += 2) {
                    push(@results, $wn->fetch($res[$i], $res[$i + 1]));
                    push(@{$morph{$res[$i + 1]}}, $res[$i]);
                }
                @results = map {node->new($_)} @results;
                foreach (keys %morph) {
                    $morph{$_} = join('|', @{$morph{$_}});
                }
                $self->{'MORPH'} = \%morph;
            } elsif ($object eq 'syno') {
                push(@results, node->new($prev->getSynset));
            } elsif ($object eq 'sfrm') {
                foreach (($prevob->framestext, $prevob->sentences)) {
                    my $frame = node->new('dummy');
                    $frame->{'EXTRA'} = $_;
                    push(@results, $frame);
                }
            }
            if (@results == 0 && $object eq 'ants'  # for indirect antonyms
                && $prev->pos eq 'a') {
                foreach ($prev->getSynset->pointers('sim')) {
                    wwSynsetRelSearch($object, $_->follow, \@results);
                }
            }
        }
    } elsif ($type eq 's') { # synset (add relations)
        foreach ($object->pointers) {
            addRel($_->type, \@results);
        }
        foreach ($object->Words) {
            foreach ($_->pointers) {
                addRel($_->type, \@results);
            }
        }
        if (scalar($object->framestext) > 0) {
            push(@results, node->new('sfrm'));
        }
        #for indirect antonyms
        if ($self->pos eq 'a' && !(grep {$_->getobject eq 'ants'} @results)){
          DOSEARCH_LOOP1:
            foreach ($object->pointers('sim')) {
                $_ = $_->follow;
                foreach ($_->Words) {
                    if (scalar($_->pointers('ants'))) {
                        push(@results, node->new('ants'));
                        last DOSEARCH_LOOP1;
                    }
                }
            }
        }
    } else { # word (add relations)
        foreach ($object->pointers) {
            addRel($_->type, \@results);
        }
        if (scalar($object->framestext) > 0) {
            push(@results, node->new('sfrm'));
        }
        #for indirect antonyms
        if ($self->pos eq 'a' && !(grep {$_->getobject eq 'ants'} @results)){
          DOSEARCH_LOOP2:
            foreach ($self->getSynset->pointers('sim')) {
                $_ = $_->follow;
                foreach ($_->Words) {
                    if (scalar($_->pointers('ants'))) {
                        push(@results, node->new('ants'));
                        last DOSEARCH_LOOP2;
                    }
                }
            }
        }
        push(@results, node->new('over'));
        push(@results, node->new('syno'));
    }
    return @results;
}


sub addRel {
    my $type = $querydata{$_[0]};
    my $resref = $_[1];
    if (!$type) { return; }
    if (!(grep { $type eq $_->getobject; } @$resref)) {
        push(@$resref, node->new($type));
    }
}


sub wwSynsetRelSearch {
    my $rel = shift;     # takes $object (relation)
    my $parent = shift;  # takes Synset object
    my $results = shift; # takes \@results array ref
    foreach my $parWord ($parent->Words) {
        if ($rel eq 'sfrm') {
            foreach my $frame ($parWord->framestext, $parWord->sentences) {
                my @set = grep {$frame eq $_->extra} @$results;
                if (@set == 0) {
                    # add frame to results, parWord to frame's RELATEDTO
                    my $temp = node->new('dummy');
                    $temp->{'EXTRA'} = $frame;
                    $temp->relatedto($parWord);
                    push(@$results, $temp);
                } else {
                    # add $parWord to $kidWord's RELATEDTO
                    $set[0]->relatedto($parWord)
                }
            }
        } else {
            foreach ($parWord->pointers($rel)) {
                my $kidWord = $_->follow;
                my @set = grep {$kidWord->id eq $_->getobject->id} @$results;
                if (@set == 0) {
                    # add kidWord to results, parWord to kidWord's RELATEDTO
                    my $temp = node->new($kidWord);
                    $temp->relatedto($parWord);
                    push(@$results, $temp);
                } else {
                    # add $parWord to $kidWord's RELATEDTO
                    $set[0]->relatedto($parWord);
                }
            }
        }
    }
}


sub unsuppress {
    my $self = shift;
    foreach (@{$self->children}) {
        $_->suppressed(0);
    }
}


sub printdebug {
    my $self = shift;
    my @kids = @{$self->children};
    if (!@kids) { return; }
    print "NODE " . $self->count . ": ";
    print scalar(@kids) . " kids";
    print " (" . join(',', map { $_->count } @kids) . ")";
    foreach (@kids) { print $_ . ', '; }
    print p;
    foreach (@kids) { $_->printdebug; }
}


1;
